/* Copyright (C) 2005 Andre Noll <maan@tuebingen.mpg.de>, see file COPYING. */

/** \file crypt_common.c Crypto functions independent of openssl/libgcrypt. */

#include <regex.h>

#include "para.h"
#include "error.h"
#include "string.h"
#include "crypt.h"
#include "crypt_backend.h"
#include "portable_io.h"
#include "fd.h"
#include "base64.h"

/** If the key begins with this text, we treat it as an ssh key. */
#define KEY_TYPE_TXT "ssh-rsa"

/*
 * Check if the given buffer starts with an ssh rsa key signature.
 *
 * Returns number of header bytes to be skipped on success, zero if no ssh rsa
 * signature was found.
 */
static size_t is_ssh_rsa_key(char *data, size_t size)
{
	char *cp;

	if (size < strlen(KEY_TYPE_TXT) + 2)
		return 0;
	cp = memchr(data, ' ', size);
	if (cp == NULL)
		return 0;
	if (strncmp(KEY_TYPE_TXT, data, strlen(KEY_TYPE_TXT)))
		return 0;
	cp++;
	if (cp >= data + size)
		return 0;
	if (*cp == '\0')
		return 0;
	return cp - data;
}

/*
 * Perform some sanity checks on the decoded ssh key.
 *
 * This function returns the size of the header. Usually, the header is 11
 * bytes long: four bytes for the length field, and the string "ssh-rsa".
 */
static int check_ssh_key_header(const unsigned char *blob, int blen)
{
	const unsigned char *p = blob, *end = blob + blen;
	uint32_t rlen;

	if (p + 4 > end)
		return -E_SSH_KEY_HEADER;
	rlen = read_u32_be(p);
	p += 4;
	if (p + rlen < p)
		return -E_SSH_KEY_HEADER;
	if (p + rlen > end)
		return -E_SSH_KEY_HEADER;
	if (rlen < strlen(KEY_TYPE_TXT))
		return -E_SSH_KEY_HEADER;
	PARA_DEBUG_LOG("type: %s, rlen: %u\n", p, rlen);
	if (strncmp((char *)p, KEY_TYPE_TXT, strlen(KEY_TYPE_TXT)))
		return -E_SSH_KEY_HEADER;
	return 4 + rlen;
}

/**
 * Perform sanity checks and base64-decode an ssh-rsa key.
 *
 * \param filename The public key file (usually id_rsa.pub).
 * \param blob Pointer to base64-decoded blob is returned here.
 * \param decoded_size The size of the decoded blob.
 *
 * The memory pointed at by the returned blob pointer has to be freed by the
 * caller.
 *
 * \return On success, the offset in bytes of the start of the key values
 * (modulus, exponent..). This is the number of bytes to skip from the blob
 * until the start of the first encoded number. On failure, a negative error
 * code is returned.
 *
 * \sa \ref uudecode().
 */
int decode_ssh_key(const char *filename, unsigned char **blob,
		size_t *decoded_size)
{
	int ret, ret2;
	void *map;
	size_t map_size;

	ret = mmap_full_file(filename, O_RDONLY, &map, &map_size, NULL);
	if (ret < 0)
		return ret;
	ret = is_ssh_rsa_key(map, map_size);
	if (ret == 0) {
		ret = -E_SSH_PARSE;
		goto unmap;
	}
	ret = uudecode(map + ret, map_size - ret, (char **)blob, decoded_size);
	if (ret < 0)
		goto unmap;
	ret = check_ssh_key_header(*blob, *decoded_size);
	if (ret < 0)
		goto unmap;
unmap:
	ret2 = para_munmap(map, map_size);
	if (ret >= 0 && ret2 < 0)
		ret = ret2;
	return ret;
}

/**
 * Check existence and permissions of a private key file.
 *
 * \param file The path of the key file.
 *
 * This checks whether the file exists and its permissions are restrictive
 * enough. It is considered an error if we own the file and it is readable for
 * others.
 *
 * \return Standard.
 */
int check_private_key_file(const char *file)
{
	struct stat st;

	if (stat(file, &st) != 0)
		return -ERRNO_TO_PARA_ERROR(errno);
	if ((st.st_uid == getuid()) && (st.st_mode & 077) != 0)
		return -E_KEY_PERM;
	return 1;
}

void hash_to_asc(unsigned char *hash, char *asc)
{
	int i;
	const char hexchar[] = "0123456789abcdef";

	for (i = 0; i < HASH_SIZE; i++) {
		asc[2 * i] = hexchar[hash[i] >> 4];
		asc[2 * i + 1] = hexchar[hash[i] & 0xf];
	}
	asc[2 * HASH_SIZE] = '\0';
}

int hash_compare(unsigned char *h1, unsigned char *h2)
{
	int i;

	for (i = 0; i < HASH_SIZE; i++) {
		if (h1[i] < h2[i])
			return -1;
		if (h1[i] > h2[i])
			return 1;
	}
	return 0;
}
